use instant::Instant;

use crate::{
    components::{monster::*, Monster, Player, Position, Renderable},
    constants::*,
    graphics::vga_color as vga,
    levels::Level,
    tile_data::{self, TileType},
};
use bracket_lib::{prelude::*, random::RandomNumberGenerator};
use hecs::{Entity, World};

pub struct Map {
    tiles: Vec<TileType>,
    tile_content: Vec<Option<Entity>>,
    border_fg: RGB,
    border_bg: RGB,
    gem_color: RGB,
}

impl BaseMap for Map {}

impl Algorithm2D for Map {
    fn dimensions(&self) -> Point {
        Point::new(MAP_WIDTH, MAP_HEIGHT)
    }
}

impl Default for Map {
    fn default() -> Self {
        Self::new()
    }
}

impl Map {
    pub fn new() -> Self {
        let mut rng = RandomNumberGenerator::new();
        Self {
            tiles: vec![TileType::Floor; MAP_SIZE],
            tile_content: vec![None; MAP_SIZE],
            border_fg: RGB::named(vga::get_by_index(rng.range(8, 15))),
            border_bg: RGB::named(vga::get_by_index(rng.range(1, 7) as usize)),
            gem_color: RGB::named(vga::get_by_index(rng.range(1, 15))),
        }
    }

    pub fn spawn_entities(&mut self, world: &mut World) {
        for (index, tile) in self
            .tiles
            .iter_mut()
            .enumerate()
            .collect::<Vec<(usize, &mut TileType)>>()
        {
            let point = Point::new(index % MAP_WIDTH, index / MAP_WIDTH);
            match tile {
                TileType::Player => {
                    let entity = world.spawn((
                        Player {
                            last_moved: Instant::now(),
                        },
                        Position {
                            x: point.x,
                            y: point.y,
                        },
                        Renderable {
                            glyph: to_cp437('☻'),
                            fg: RGB::named(vga::YELLOW_BRIGHT),
                            bg: RGB::named(vga::BLACK),
                        },
                    ));
                    self.tile_content[index] = Some(entity);
                }
                TileType::Slow => {
                    let mut rng = RandomNumberGenerator::new();
                    let entity = world.spawn((
                        Monster {
                            kind: MonsterKind::Slow,
                            ticks_until_move: ticks_for_kind(MonsterKind::Slow),
                        },
                        Position {
                            x: point.x,
                            y: point.y,
                        },
                        Renderable {
                            glyph: *rng
                                .random_slice_entry(&glyphs_for_kind(MonsterKind::Slow))
                                .unwrap(),
                            fg: color_for_kind(MonsterKind::Slow),
                            bg: RGB::named(vga::BLACK),
                        },
                    ));
                    self.tile_content[index] = Some(entity);
                }
                TileType::Medium => {
                    let mut rng = RandomNumberGenerator::new();
                    let entity = world.spawn((
                        Monster {
                            kind: MonsterKind::Medium,
                            ticks_until_move: ticks_for_kind(MonsterKind::Medium),
                        },
                        Position {
                            x: point.x,
                            y: point.y,
                        },
                        Renderable {
                            glyph: *rng
                                .random_slice_entry(&glyphs_for_kind(MonsterKind::Medium))
                                .unwrap(),
                            fg: color_for_kind(MonsterKind::Medium),
                            bg: RGB::named(vga::BLACK),
                        },
                    ));
                    self.tile_content[index] = Some(entity);
                }
                TileType::Fast => {
                    let mut rng = RandomNumberGenerator::new();
                    let entity = world.spawn((
                        Monster {
                            kind: MonsterKind::Fast,
                            ticks_until_move: ticks_for_kind(MonsterKind::Fast),
                        },
                        Position {
                            x: point.x,
                            y: point.y,
                        },
                        Renderable {
                            glyph: *rng
                                .random_slice_entry(&glyphs_for_kind(MonsterKind::Fast))
                                .unwrap(),
                            fg: color_for_kind(MonsterKind::Fast),
                            bg: RGB::named(vga::BLACK),
                        },
                    ));
                    self.tile_content[index] = Some(entity);
                }
                _ => {}
            }
        }
    }

    pub fn clear_all_tile_content(&mut self) {
        self.tile_content.iter_mut().for_each(|tile| {
            *tile = None;
        });
    }

    pub fn get_tile_content_at(&self, point: Point) -> Option<Entity> {
        let index = self.point2d_to_index(point);
        self.tile_content[index]
    }

    pub fn set_tile_content_at(&mut self, point: Point, entity: Entity) {
        let index = self.point2d_to_index(point);
        self.tile_content[index] = Some(entity);
    }

    pub fn clear_tile_content_at(&mut self, point: Point) {
        let index = self.point2d_to_index(point);
        self.tile_content[index] = None;
    }

    pub fn get_tiles(&self) -> &Vec<TileType> {
        &self.tiles
    }

    pub fn draw(&self, bterm: &mut BTerm) {
        // Border
        bterm.fill_region(
            Rect {
                x1: MAP_X as i32 - 1,
                x2: MAP_WIDTH as i32 + MAP_X as i32 + 1,
                y1: MAP_Y as i32 - 1,
                y2: MAP_HEIGHT as i32 + MAP_Y as i32 + 1,
            },
            to_cp437('▓'),
            self.border_fg,
            self.border_bg,
        );

        for (i, tile) in self.tiles.iter().enumerate() {
            let point = self.index_to_point2d(i);
            let (x, y) = (point.x as usize, point.y as usize);

            let data = tile_data::tile_data(*tile);
            bterm.set(
                x + MAP_X,
                y + MAP_Y,
                if *tile == TileType::Gem {
                    self.gem_color
                } else {
                    RGB::named(data.color_fg)
                },
                data.color_bg,
                data.glyph,
            );
        }
    }

    pub fn get_tile_at(&self, point: Point) -> TileType {
        self.tiles[self.point2d_to_index(point)]
    }

    pub fn get_tile_at_mut(&mut self, point: Point) -> &mut TileType {
        let index = self.point2d_to_index(point);
        &mut self.tiles[index]
    }

    pub fn set_tile_at(&mut self, point: Point, tile: TileType) {
        let index = self.point2d_to_index(point);
        self.tiles[index] = tile;
    }
}

impl From<Level> for Map {
    fn from(level: Level) -> Self {
        match level {
            Level::Normal(data) => {
                assert_eq!(data.len(), MAP_HEIGHT, "Level data is incorrect height!");

                let mut map = Self::new();
                for (y, line) in data.iter().enumerate() {
                    assert_eq!(line.len(), MAP_WIDTH, "Level data is incorrect width!");
                    for (x, c) in line.chars().enumerate() {
                        map.set_tile_at(
                            Point {
                                x: x as i32,
                                y: y as i32,
                            },
                            TileType::from(c),
                        );
                    }
                }
                map
            }
            Level::Randomized(data) => {
                let mut rng = RandomNumberGenerator::new();
                let mut map = Self::new();
                for (tile, count) in data {
                    for _ in 0..count {
                        loop {
                            let point = Point {
                                x: rng.range(0, MAP_WIDTH as i32),
                                y: rng.range(0, MAP_HEIGHT as i32),
                            };
                            if map.get_tile_at(point) == TileType::Floor {
                                map.set_tile_at(point, tile);
                                break;
                            }
                        }
                    }
                }
                loop {
                    let point = Point {
                        x: rng.range(0, MAP_WIDTH as i32),
                        y: rng.range(0, MAP_HEIGHT as i32),
                    };
                    if map.get_tile_at(point) == TileType::Floor {
                        map.set_tile_at(point, TileType::Player);
                        break;
                    }
                }
                map
            }
            Level::End => todo!(),
        }
    }
}
