use instant::Instant;

use crate::resources::clock::Clock;
use crate::resources::flashing_message::{FlashingMessage, FlashingMessageIntent};
use crate::resources::sound_effects::SoundEffects;
use crate::resources::stats::Stats;
use crate::resources::{difficulty::*, MessageShown};
use crate::resources::{Map, Resources, SoundOutput};
use crate::{graphics, input, levels, systems};
use bracket_lib::prelude::*;
use hecs::World;

pub struct State {
    world: World,
    resources: Resources,
}

impl GameState for State {
    fn tick(&mut self, bterm: &mut BTerm) {
        input::handle(&mut self.world, &mut self.resources, bterm);
        systems::run(&mut self.world, &mut self.resources);
        graphics::draw(&self.world, &mut self.resources, bterm);

        if self.resources.should_advance_level {
            self.next_level();
        }
    }
}

impl State {
    #[allow(dead_code)]
    pub fn new(initialize_sound: bool) -> Self {
        let mut sound_output = if initialize_sound {
            Some(SoundOutput::new())
        } else {
            None
        };

        let sound_effects = sound_output.as_ref().map(SoundEffects::new);

        if let (Some(so), Some(se)) = (&mut sound_output, &sound_effects) {
            so.play_sound(se.startup.clone());
        }

        let starting_level = 0;
        let selected_difficulty = Difficulty::Secret;
        let difficulty_settings = DifficultySettings::from(&selected_difficulty);

        let mut world = World::new();

        let mut map = Map::from(levels::get_level(starting_level));
        map.spawn_entities(&mut world);

        let resources = Resources {
            level_number: starting_level,
            show_debug_info: false,
            player_input: None,
            stats: Stats {
                score: 0,
                gems: difficulty_settings.starting_gems,
                whips: difficulty_settings.starting_whips,
                whip_power: difficulty_settings.starting_whip_power,
                teleports: difficulty_settings.starting_teleports,
                keys: difficulty_settings.starting_keys,
            },
            clock: Clock {
                last_ticked: Instant::now(),
                has_ticked: false,
                ticks: 0,
            },
            stop_clock: false,
            map,
            sound_effects,
            sound_output,
            difficulty: Some(difficulty_settings),
            flashing_message: Some(FlashingMessage::new(
                "Press any key to begin this level.",
                Some(FlashingMessageIntent::Start),
            )),
            message_shown: match selected_difficulty {
                Difficulty::Novice | Difficulty::Experienced => Default::default(),
                _ => MessageShown::all_shown(),
            },
            should_advance_level: false,
            speed_time_ticks: 0,
            slow_time_ticks: 0,
        };

        State { world, resources }
    }

    fn next_level(&mut self) {
        self.resources.level_number += 1;

        self.world.clear();
        self.resources.map = Map::from(levels::get_level(self.resources.level_number));
        self.resources.map.spawn_entities(&mut self.world);

        self.resources.should_advance_level = false;
    }
}
